\ compiler.fs
\ Hannes Hauer
\ Simon Himmelbauer
\ Siegfried Oleg Pammer
\ Description: Chip-8-to-Forth-compiler

needs random.fs

needs ../common/decoder.fs
needs ../common/constants.fs

\ Utility words
\ Print hex-stack item on seperate line
: trace ( n -- )
  hex. s\" \n" type
;

\ CHIP-8 register handling
: register8 ( n "name" -- )
  create ,
  does> ( -- n )
      ( addr ) @ $FF and ;

: register16 ( n "name" -- )
  create ,
  does> ( -- n )
      ( addr ) @ ;

0 register16 __regi 
0 register8 __reg0
0 register8 __reg1
0 register8 __reg2
0 register8 __reg3
0 register8 __reg4
0 register8 __reg5
0 register8 __reg6
0 register8 __reg7
0 register8 __reg8
0 register8 __reg9
0 register8 __rega
0 register8 __regb
0 register8 __regc
0 register8 __regd
0 register8 __rege
0 register8 __regf
0 variable __delay drop
0 variable __sound drop 

create registers
    ' __reg0 , ' __reg1 , ' __reg2 , ' __reg3 , 
    ' __reg4 , ' __reg5 , ' __reg6 , ' __reg7 , 
    ' __reg8 , ' __reg9 , ' __rega , ' __regb , 
    ' __regc , ' __regd , ' __rege , ' __regf , 
    ' __regi ,

\ Used to retrieve the address of a register via >body
: int-to-reg ( n -- addr )
    dup assert( 17 < )
    cells registers + @
;

\ Load value from register
: int-to-load ( n -- addr )
    dup assert( 17 < )
    cells registers + @ postpone literal postpone execute
;

\ CHIP-8 subroutine handling
create gotos 2048 cells allot
: mark_goto here swap 1- cells gotos + ! ; immediate
: goto r> drop 1- cells gotos + @ >r ;

\ The function mapper stores all jump and call targets
create function-mapping 2048 cells allot
0 variable function-mapping-count drop

\ Retrieve the goto target value for the given address n
\ Returns -1 in case no such address could be found
: get-call-target { n -- n }
  \ In case target is not found
  -1
  function-mapping-count @ 0 ?do
    function-mapping i cells + @
    \ Wrong values
    n = if
      drop
      i 1+
      leave
    endif
  loop
;

\ This area is used to handle redundant key presses
\ The issue is that some CHIP-8 instructions check whether a key is pressed
\ but a repeating key press is not desired to be handled
\ Therefore, we store the following states
\ 0: Key was not pressed
\ 1: Key was pressed but not yet consumed
\ 2: Key was pressed and consumed
create prev-key-state 0 , 0 , 0 , 0 ,
                        0 , 0 , 0 , 0 ,
                        0 , 0 , 0 , 0 ,
                        0 , 0 , 0 , 0 ,

\ Update key-state table
: run-update-key-state ( -- )
  sdl-pumpevents
  NULL sdl-getkeyboardstate
  16 0 ?do
    i cells key-table + @
    over + c@
    
    i cells prev-key-state + @
    case
      0 of i cells prev-key-state + ! endof
      1 of    
        0 = if
          0 i cells prev-key-state + !
        endif
      endof
      2 of
        0 = if
          0 i cells prev-key-state + !
        endif
      endof
      swap drop
    endcase
  loop
  drop
;

\ Consume the key-state which means it cannot be reused
: run-consume-key-state ( n -- )
  2 swap cells prev-key-state + !
;
 
\ Retrieve key-state. Only if state is 1, the keypress will be recognized
: run-get-key-state ( n -- flag )
  \ Retrieve corresponding SCANCODE
  run-update-key-state
  dup cells prev-key-state + @ 1 = if
    run-consume-key-state
    true
  else
    drop
    false
  endif
;

\ Retrieve the current time and convert it to an integer
: utime-int ( -- ntime )
  utime d>s
;

create last-measure 0 ,
: run-reset-measurement ( -- )
  utime-int last-measure !
;

: run-decrement-timer ( -- )
  \ Make sure timer is decremented only 60 times per second
  __delay @ 0> if
    \ Calculate how much time in microseconds has passed since the last decrement of the timer
    utime-int dup last-measure @ - \ Timediff
    \ Check whether there is time left
    \ 16666 equals to 1 / 60 in microseconds which refers to the time frame per instruction
    16666 -
    0> if
      \ more than 1 / 60 of a second has passed, decrement timer
      __delay @ 1- __delay !
      \ Store currently measured time
      last-measure !
    else
        drop
    endif
  endif
;

create event c-sizeof-event allot
\ Management function to keep SDL happy and handle events
: run-poll-event ( -- )
  run-decrement-timer
  event sdl-pollevent 0 > if
    event 0 + ul@
    SDL_QUIT = if bye endif
  endif
;

\ Wait until a key has been pressed
: run-get-key-pressed ( -- key )
  begin
  16 0 ?do
    \ Load scancode
    i cells prev-key-state + @
    1 = if
      i
      dup run-consume-key-state
      unloop
      exit 
    endif
  loop
  run-update-key-state
  run-poll-event
  again
;

: run-get-bcd-digit { n index -- digit }
  1 index 0 ?do 10 * loop \ pow(10, index)
  n swap
  / 10 mod
;

: run-store-bcd-digits { n -- }
  __regi $200 - image + 
  dup 0 + n 2 run-get-bcd-digit swap c!
  dup 1 + n 1 run-get-bcd-digit swap c!
  2 + n 0 run-get-bcd-digit swap c!
;

\ CHIP-8 instructions
\ 00E0
: run-clear-screen
  black fill-surface
  update-surface
;

\ 3XNN / 4XNN
: compile-cmp-imm { imm target-reg -- }
  flag
  case
    FLAG_EQUAL of
      imm postpone literal target-reg int-to-load postpone =
    endof
    FLAG_NEQUAL of
      imm postpone literal target-reg int-to-load postpone <>
    endof 
    s\" Unknown flag" type
  endcase
;

\ EX9E / EXA1
: compile-cmp-key { target-reg -- }
  flag
  case
    FLAG_EQUAL of
      target-reg int-to-load postpone run-get-key-state
    endof
    FLAG_NEQUAL of
      target-reg int-to-load postpone run-get-key-state postpone invert
    endof 
    s\" Unknown flag" type
  endcase
;

\ 5XY0 / 9XY0
: compile-cmp-reg { reg-x reg-y -- }
  flag
  case
    FLAG_EQUAL of
      reg-x int-to-load reg-y int-to-load postpone =
    endof
    FLAG_NEQUAL of
      reg-x int-to-load reg-y int-to-load postpone <>
    endof 
    s\" Unknown flag" type
  endcase
;

\ DXYN
: run-draw-screen { x y imm -- }
  0 [ ' __regf >body ]L ! \ Init VF to 0
  x 64 mod \ x
  y 64 mod \ x y
  dup 32 >= if
    0
  else
    dup imm + 32 >= if
      dup 32 swap -
    else
      imm
    endif
  endif
  \ Stack effect: x y imm
  0 ?do
    2dup i + \ x y x (y + i)
    __regi $ffff > if
      __regi i +
    else
      __regi $200 - i + image +
    endif
    
    \ x y x (y + i) adr_i
    c@ rot \ x y (y + i) c x
    swap align-byte \ x y (y + i) c
    swap cells chp8-screen + dup @ \ x y c adr_screen c2
    rot 2dup \ x y adr_screen c2 c c2 c
    xor \ x y adr_screen c2 c (c^c2)
    -rot \ x y adr_screen (c^c2) c2 c
    and 0<> \ x y adr_screen (c^c2) flag
    if
      1 [ ' __regf >body ]L ! \ Init VF to 0
    endif
    \ x y adr_screen (c^c2)
    swap ! \ x y
    \ x y
  loop
  2drop
  update-surface
;

\ FX0A
: compile-get-key { target-reg -- }
  postpone run-get-key-pressed target-reg int-to-reg >body postpone literal postpone !
  target-reg int-to-load postpone hex.
;

\ FX07
: compile-get-time { target-reg -- }
  __delay postpone literal postpone @
  target-reg int-to-reg >body postpone literal postpone !
;

\ ANNN / 6XNN
: compile-imm { imm target-reg -- }
  imm postpone literal target-reg int-to-reg >body postpone literal postpone !
;

\ 7XNN
: compile-add { imm target-reg -- }
  target-reg int-to-load imm postpone literal postpone + 
  target-reg int-to-reg >body postpone literal postpone !
;

\ 6XNN / 7XNN
\ VX = NN / VX += NN
: compile-ar { imm target-reg -- }
  flag
  case
    FLAG_NONE of imm target-reg compile-imm endof
    FLAG_ADD of imm target-reg compile-add endof
    s\" Unknown flag" type
  endcase
;

\ 8XYf: Logic / Arithmetic
: compile-mov-reg { target-reg source-reg -- }
  flag
  case
    FLAG_NONE of \ 8XY0: VX = VY
      source-reg int-to-load 
    endof
    FLAG_BIT_OR of \ 8XY1: VX |= VY
      target-reg int-to-load source-reg int-to-load postpone or
    endof
    FLAG_BIT_AND of \ 8XY2: VX &= VY
      target-reg int-to-load source-reg int-to-load postpone and
    endof
    FLAG_BIT_XOR of \ 8XY3: VX ^= VY
      target-reg int-to-load source-reg int-to-load postpone xor
    endof
    FLAG_ADD of \ 8XY4: VX += VY
      source-reg int-to-load target-reg int-to-load postpone + postpone dup \ add
      255 postpone literal postpone > 1 postpone literal postpone and     \ check overflow
      15 int-to-reg >body postpone literal postpone !                     \ Store overflow flag in VF
    endof
    FLAG_SUB of \ 8XY5: VX -= VY, VF = VX > VY
      target-reg int-to-load source-reg int-to-load postpone 2dup 
      postpone > 1 postpone literal postpone and                  \ Check underflow
      15 int-to-reg >body postpone literal postpone !             \ Store underflow flag in VF
      postpone - 
    endof
    FLAG_RSHIFT of \ 8XY6: VF = VX-LSB, VX >> 1
      target-reg int-to-load postpone dup
      1 postpone literal postpone and                 \ bit value to be shifted out
      15 int-to-reg >body postpone literal postpone ! \ ... stored in VF
      1 postpone literal postpone rshift              \ VX right-shifted by one
    endof
    FLAG_SUB_INV of \ 8XY7: VX = VY - VX, VF = VY > VX
      source-reg int-to-load target-reg int-to-load postpone 2dup 
      postpone > 1 postpone literal postpone and                  \ Check underflow
      15 int-to-reg >body postpone literal postpone !             \ Store underflow flag in VF
      postpone - 
    endof
    FLAG_LSHIFT of \ 8XYE: VF = VX-MSB, VX << 1;
      target-reg int-to-load postpone dup
      $80 postpone literal postpone and 7 postpone literal postpone rshift    \ bit value to be shifted out
      15 int-to-reg >body postpone literal postpone !                         \ ...stored into VF
      1 postpone literal postpone lshift                                      \ VX left-shifted by one
    endof
    s\" Unknown flag " type flag .
  endcase
  target-reg int-to-reg >body postpone literal postpone ! \ Store result in VX
;

\ FX1E / FX33
: compile-mov-reg-i { source-reg -- }
  flag
  case
    FLAG_ADD of \ FX1E: I += VX
      16 int-to-load source-reg int-to-load postpone + $FFFF postpone literal postpone and
      16 int-to-reg >body postpone literal postpone !
    endof
    FLAG_BCD of \ FX33: I = VX digit 1, I+1 = VX digit 2, I+2 = VX digit 3
      source-reg int-to-load postpone run-store-bcd-digits           
    endof
    s\" Unknown flag" type
  endcase
;

\ CXNN
: compile-rand { imm target-reg -- }
  256 postpone literal postpone random imm postpone literal postpone and
  target-reg int-to-reg >body postpone literal postpone !
;

\ FX55: Copy registers 0 to X to address I
: compile-reg-dump { final-reg -- }
  final-reg 1+ 0 ?do
    \ First, we load the address from register i subtract the base address of $200 and add
    \ the base address of our emulated memory
    i int-to-load 16 int-to-load image $200 - postpone literal postpone + 
    \ Add offset and store value
    i postpone literal postpone + postpone c!
  loop
;

\ FX65: Copy from address I to registers 0 to X
: compile-reg-load { final-reg -- }
  final-reg 1+ 0 ?do
    \ First, we load the address from register i subtract the base address of $200 and add
    \ the base address of our emulated memory
    postpone __regi image $200 - postpone literal postpone +
    \ Add offset and load value
    i postpone literal postpone + postpone c@ 
    i int-to-reg >body postpone literal postpone !
  loop
;

\ FX29: Set register I to address of character X
: compile-set-sprite { source-reg -- }
  sprites postpone literal source-reg int-to-load 5 postpone literal postpone * postpone + 
  16 int-to-reg >body postpone literal postpone !
;

\ FX15 / FX18
: compile-set-time { target-reg -- }
  flag
  case
    FLAG_SOUND of
      target-reg int-to-load
      __sound postpone literal postpone !
    endof
    FLAG_DELAY of
      target-reg int-to-load
      __delay postpone literal postpone !
      postpone run-reset-measurement
    endof
  endcase
;

\ Compiler

defer program
: compile-call { target }
  \ Make recursive call with appropriate jump argument
  target get-call-target
  postpone literal postpone program
;

\ Compile CHIP-8 instruction into Forth instructions
: compile-chp8 { opcode address -- }
  \ Check whether current location requires goto-marker
  address
  get-call-target dup -1 <> if
    s\" Inserting goto-marker\n" type
    postpone mark_goto
  else
    drop
  endif
  
  \ Inject runtime code
  postpone run-poll-event
            
  address postpone literal postpone trace \ Debug-print CHIP-8 PC
  
  opcode
  case
    INST_CALL of imm compile-call endof
    INST_CLEAR of postpone run-clear-screen endof
    INST_CMP_IMM of 
      imm register1 compile-cmp-imm
      postpone if
      address 4 + get-call-target postpone literal postpone goto
      postpone endif
    endof
    INST_CMP_KEY of 
      register1 compile-cmp-key
      postpone if
      address 4 + get-call-target postpone literal postpone goto
      postpone endif
    endof
    INST_CMP_REG of 
      register1 register2 compile-cmp-reg
      postpone if
      address 4 + get-call-target postpone literal postpone goto
      postpone endif
    endof
    INST_DRAW of 
      register1 int-to-load
      register2 int-to-load
      imm postpone literal
      postpone run-draw-screen  
    endof
    INST_GET_KEY of register1 compile-get-key endof
    INST_GET_TIME of register1 compile-get-time endof
    INST_GOTO of imm get-call-target postpone literal postpone goto endof
    INST_MOV_IMM of imm register1 compile-ar endof
    INST_MOV_IMM_I of imm 16 compile-imm endof
    INST_MOV_REG of register1 register2 compile-mov-reg endof
    INST_MOV_REG_I of register1 compile-mov-reg-i endof
    INST_NOP of endof
    INST_RAND of imm register1 compile-rand endof
    INST_REG_DUMP of register1 compile-reg-dump endof
    INST_REG_LOAD of register1 compile-reg-load endof
    INST_RETURN of postpone exit endof
    INST_SET_SPRITE of register1 compile-set-sprite endof
    INST_SET_TIME of register1 compile-set-time endof
    s\" Unknown opcode\n" type
  endcase
  0
;

\ Decode program step by step and invoke xt with the given instruction flag
: prepare-program { xt -- }
  codesize 0 +do
    image i + uw@ \ Instruction
    i $200 + hex. s" " type dup hex. s\" \n" type \ Debug-print instruction address and opcode
    \ separate into nibbles
    dup $f and to nibble2 4 rshift
    dup $f and to nibble1 4 rshift
    dup $f and to nibble4 4 rshift
    $f and to nibble3
    decode-instruction
    $200 i + xt execute
    -1 = if
      leave
    endif
  2 +loop
;

\ Call trampoline to implement the call instruction
\ If the chip 8 program is called with an argument != 0,
\ then goto jumps to the appropriate place
: compile-call-trampoline ( -- )
  postpone dup postpone 0> 
  postpone if postpone goto
  postpone else postpone drop
  postpone endif
;

\ Stores the address of all targets used by Jump, Call or Skip instructions
: jump-analysis { opcode address -- }
  opcode -1 <> if
    opcode INST_GOTO = opcode INST_CALL = + 0< if
      \ Only store if address is not yet a target
      imm get-call-target -1 = if
        \ Store the target in the function mapper
        imm function-mapping function-mapping-count @ cells + !
        function-mapping-count @ 1+ function-mapping-count !
      endif
    endif
    opcode INST_CMP_REG = opcode INST_CMP_IMM = + opcode INST_CMP_KEY = + 0< if
      \ Only store if address is not yet a target
      address 4 + get-call-target -1 = if
        \ Store the address of the instruction after the next instruction
        address 4 + function-mapping function-mapping-count @ cells + !
        function-mapping-count @ 1+ function-mapping-count !
      endif
    endif 
  endif
  0
;

\ This program contains the compile CHIP-8 program
:noname ( n -- )
  [ 
    \ Load image
    next-arg load-image
    \ Insert trampoline
    compile-call-trampoline
    \ Run Jump Analysis
    ' jump-analysis prepare-program
    s\" \nJUMP ANALYSIS COMPLETED\n" type
    \ Compile Chip-8 Code
    ' compile-chp8 prepare-program
    ] 
;
is program
