\ executor.fs
\ Hannes Hauer
\ Simon Himmelbauer
\ Siegfried Oleg Pammer
\ Description: Chip-8 interpreter

needs random.fs

needs ../common/util.fs
needs ../common/decoder.fs
needs ../common/constants.fs
needs regfile.fs

\ Utility words
\ Add registers to stack
: push-regs 
  register1 chp8-get-reg register2 chp8-get-reg
;

: coerce-hex { input -- hex or -1 }
  \ converts an SDL_Scancode value into 0-F, if hex,
  \ otherwise returns -1
  16 0 ?do
    key-table i cells + @ input = if i unloop exit endif
  loop
  -1 exit
;

: get-key-state 
  NULL sdl-getkeyboardstate \ do NOT free the returned memory!
  \ dup . s\" get-key-state\n" type
  16 0 ?do
    dup key-table i cells + @ + c@ 1 = if drop i unloop exit endif
  loop
  drop -1 exit
;

: align-byte { offset c -- c }
  64 8 - offset - \ offset 
  dup 0< if 
    negate c swap rshift
  else
    c swap lshift
  endif
;

: get-bcd-digit { index -- digit }
  1 index 0 ?do 10 * loop \ pow(10, index)
  register1 chp8-get-reg swap
  / 10 mod
;

\ CHIP-8 instructions
: chp8-nop
  0
;

\ 00E0
: chp8-clear-disp
  black fill-surface
  update-surface
  0 \ return 0
;

\ 3XNN / 4XNN
: chp8-cmp-imm
  flag
  case
    FLAG_EQUAL of
      imm register1 chp8-get-reg =
    endof
    FLAG_NEQUAL of 
      imm register1 chp8-get-reg <>
    endof
    s\" chp8-cmp-imm: Unknown flag\n" chp8-error exit
  endcase
  assert0( depth 1 = )
  if
    pc 2 + to pc
  endif
  0 \ return 0
;

\ EX9E / EXA1
: chp8-cmp-key \ if (Vx op get_key()) skip next;
  flag
  case
    FLAG_EQUAL of
      get-key-state
      register1 chp8-get-reg
      =
    endof
    FLAG_NEQUAL of 
      get-key-state
      register1 chp8-get-reg
      <>
    endof
    s\" chp8-cmp-key: Unknown flag\n" chp8-error exit
  endcase
  assert0( depth 1 = )
  if
    pc 2 + to pc
  endif
  0 \ return 0
;

\ 5XY0 / 9XY0
: chp8-cmp-reg \ if (Vx op Vy) skip next;
  flag
  case
    FLAG_EQUAL of
      push-regs =
    endof
    FLAG_NEQUAL of 
      push-regs <>
    endof
    s\" chp8-cmp-reg: Unknown flag\n" chp8-error exit
  endcase
  assert0( depth 1 = )
  if
    pc 2 + to pc
  endif
  0 \ return 0
;

\ DXYN
: chp8-draw \ draw(Vx, Vy, N)
  15 0 chp8-set-reg \ Init VF to 0
  register1 chp8-get-reg 64 mod \ x 
  register2 chp8-get-reg 64 mod \ y
  dup 32 >= if
    0
  else
    dup imm + 32 >= if
      dup 32 swap -
    else
      imm
    endif
  endif
  0 ?do
    2dup i + \ x y x (y + i) 
    chp8-reg-i $ffff > if
      chp8-reg-i i +
    else
      chp8-reg-i $200 - i + image +
    endif
    \ x y x (y + i) adr_i
    c@ rot \ x y (y + i) c x
    swap align-byte \ x y (y + i) c
    swap cells chp8-screen + dup @ \ x y c adr_screen c2
    rot 2dup \ x y adr_screen c2 c c2 c
    xor \ x y adr_screen c2 c (c^c2)
    -rot \ x y adr_screen (c^c2) c2 c
    and 0<> \ x y adr_screen (c^c2) flag
    if 
      15 1 chp8-set-reg
    endif
    \ x y adr_screen (c^c2)
    swap ! \ x y
    \ x y
  loop
  2drop
  update-surface
  0 \ return 0
;

\ FX0a
: chp8-get-key
  c-sizeof-event allocate assert0( 0 = )
  begin
    dup sdl-waitevent 0 = if
      sdl-geterror type
      s\" \n" type
      free assert0( 0 = )
      -1 exit
    endif
    dup 0 + ul@
    case
      SDL_QUIT of
        codesize to pc
        free assert0( 0 = )
        0 exit
      endof
      SDL_KEYDOWN of
        dup c-offsetof-kbdevt-repeat + c@
        0 = if
          dup c-offsetof-kbdevt-keysym + ul@
          coerce-hex dup -1 <> if
            register1 swap chp8-set-reg
            free assert0( 0 = )
            0 exit
          endif
          drop
        endif
      endof
    endcase
  again
;

\ FX07
: chp8-get-time
  register1 chp8-delay-timer f@ f>s chp8-set-reg
  0
;

\ 1NNN
: chp8-goto
  imm $200 - to pc
  0 \ return 0
;

\ BNNN
: chp8-jump \ Flow
    0 chp8-get-reg \ v0
    imm $200 - + \ (v0+imm)
    to pc
    0 \ return 0
;

\ 6XNN / 7XNN
: chp8-mov-imm \ Vx (+)= NN
  flag
  case
    FLAG_NONE of
      register1 imm chp8-set-reg
    endof
    FLAG_ADD of
      register1 dup \ register1 2x on stack (for later)
      chp8-get-reg \ get current value from register1
      imm +  \ add imm
      chp8-set-reg \ store new value in register1
    endof
    s\" chp8-mov-imm: Unknown flag\n" chp8-error exit
  endcase
  0 \ return 0
;

\ ANNN
: chp8-mov-imm-i \ I = NNN
  imm to chp8-reg-i
  0 \ return 0
;

\ 8XY0
: chp8-mov-reg \ Vx op= Vy
  flag
  case
    FLAG_NONE    of register2 chp8-get-reg endof
    FLAG_BIT_OR  of push-regs or endof
    FLAG_BIT_AND of push-regs and endof
    FLAG_BIT_XOR of push-regs xor endof
    FLAG_ADD     of
      push-regs +
      dup 255 > 1 and 15 swap chp8-set-reg
    endof
    FLAG_SUB     of
      push-regs \ a b
      2dup > 1 and 15 swap chp8-set-reg \ set VF if a > b
      - \ a - b
    endof
    FLAG_RSHIFT  of 
      register1 chp8-get-reg 
      \ Extract least significant bit and set to VF
      dup $1 and 15 swap chp8-set-reg
      \ Apply shift right by 1
      1 rshift
    endof
    FLAG_SUB_INV of
      register2 chp8-get-reg register1 chp8-get-reg \ a b
      2dup > 1 and 15 swap chp8-set-reg \ set VF if a > b
      - \ a - b
    endof
    FLAG_LSHIFT  of 
      register1 chp8-get-reg 
      \ Extract most significant bit and set to VF
      dup $80 and \ reg msb
      7 rshift \ reg msb
      15 swap chp8-set-reg
      \ Apply shift left by 1
      1 lshift
    endof
    s\" chp8-mov-reg: Unknown flag\n" chp8-error exit
  endcase
  register1 swap chp8-set-reg
  0 \ return 0
;

\ FX1E / FX33
: chp8-mov-reg-i
  flag
  case
    FLAG_ADD of \ I += Vx
      \ Note: VF should not be set.
      chp8-reg-i
      register1 chp8-get-reg
      + to chp8-reg-i
    endof
    FLAG_BCD of
      chp8-reg-i $200 - image +
      dup 0 + 2 get-bcd-digit swap c!
      dup 1 + 1 get-bcd-digit swap c!
          2 + 0 get-bcd-digit swap c!
    endof
    s\" chp8-mov-reg-i: Unknown flag\n" chp8-error exit
  endcase
  0 \ return 0
;

\ CXNN
: chp8-rand \ Vx = rand() & NN;
  register1 256 random imm and chp8-set-reg
  0
;

\ FX55
: chp8-reg-dump \ reg_dump(Vx, &I)
  chp8-reg
  chp8-reg-i $200 - image +
  register1 1+
  move
  0
;

\ FX65
: chp8-reg-load \ reg_load(Vx, &I)
  chp8-reg-i $200 - image +
  chp8-reg
  register1 1+
  move
  0
;

\ 00EE
: chp8-return 
  pop-return to pc 
  0 
;

\ FX29
: chp8-set-sprite
  sprites register1 chp8-get-reg 5 * + to chp8-reg-i
  0
;

\ FX15 / FX18
: chp8-set-time
  register1 chp8-get-reg s>f
  flag
  case
    FLAG_SOUND of
       chp8-sound-timer f!
    endof
    FLAG_DELAY of
       chp8-delay-timer f!
    endof
    fdrop s\" chp8-set-time: Unknown flag\n" chp8-error exit
  endcase
  0
;

\ 2NNN
: chp8-call \ goto NNN
  pc push-return
  imm $200 - to pc
  0 \ return 0
;

\ Interpreter
create instructions-table
  ' chp8-nop ,        ' chp8-clear-disp , ' chp8-cmp-imm , 
  ' chp8-cmp-key ,    ' chp8-cmp-reg ,    ' chp8-draw , 
  ' chp8-get-key ,    ' chp8-get-time ,   ' chp8-goto , 
  ' chp8-jump ,       ' chp8-mov-imm ,    ' chp8-mov-imm-i , 
  ' chp8-mov-reg ,    ' chp8-mov-reg-i ,  ' chp8-rand , 
  ' chp8-reg-dump ,   ' chp8-reg-load ,   ' chp8-return , 
  ' chp8-set-sprite , ' chp8-set-time ,   ' chp8-call ,

: execute-instruction { opcode -- }
  opcode cells instructions-table + perform
;
